      subroutine memdof(nnode,nnodt,nelem,npoin,ndofn,ntotv,lpntn,
     .                  lnods,svomt,svome,speme,sizms,lengp,iblok,
     .                  nzeco,p_amatr,p_roffs,p_acolu) 
c************************************************************************
c                 
c**** This routine allocates memory for the coefficients of a matrix A
c**** stored in compressed sparse row (CSR) format. This matrix is
c**** assumed to come from a problem of the form     
c
c     A_f x_f = F - A_p x_p
c
c**** where x_f is unknown and x_p are prescribed values. Moreover, the
c**** vector x is allowed to be of the form x = ( y , z ) (when it is 
c**** properly ordered), where y are referred to as the `main' degrees
c**** of freedom and z is thought of as a Lagrange multiplier.
c
c**** INPUT:
c      
c     nnode:   Number of nodes per element of y
c     nnodt:   Total number of nodes per element (of y and z)
c     nelem:   Number of elements            
c     npoin:   Number of nodal points of y
c     ndofn:   Number of degrees of freedom per node of y (ONLY 1 for z) 
c     ntotv:   Total number of degrees of freedom (y and z)            
c     lpntn:   Array of ordering of d.o.f. to define prescriptions and
c              to improve the efficiency (if wanted). It is assumed that
c              lpntn(i) > 0 if i is free and < 0 if i is prescribed
c     lnods:   Matrix of nodal connections (of both y and z. It has
c              dimension (nnodt,nelem))
c     svomt:   Counter for volatile memory
c     svome:   Maximum of volatile memory
c     speme:   Counter for permanent memory
c     sizms:   Size previously needed to store A and the arrays of indexes
c     lengp:   Length of integers (4 or 8)
c     iblok:   Code for the contruction of matrices, 1 for A_f, 2 for A_p
c
c**** OUTPUT:
c
c     nzeco:   Number of nonzero coefficients of the matrix constructed
c     p_amatr: Pointer to matrix A (amatr(nzeco))
c     p_roffs: Pointer to the array of rows roffs(ntotv+1) (roffs(i) = 
c              coefficient of A where eqn. i starts)
c     p_acolu: Pointer to the array of columns acolu(nzeco) (acolu (i)
c              = column of the i coefficient of matrix A)
c                 
c************************************************************************
      implicit none
      integer   nnode,nnodt,nelem,npoin,ndofn,ntotv,svomt,
     .          svome,speme,sizms,lengp,iblok,nzeco
      integer   p_amatr,p_roffs,p_acolu                     ! pointers
      integer   lpntn(ntotv),lnods(nnodt,nelem) 
      integer   nlist,ncoef,itotv,itotn,ipoin,kflag,idofn,
     .          lbyts,izeco,npoit,nnodw
      integer   p_nepoi,p_lelpo,p_touch,iadd1,iadd2         ! pointers
c
c*** Initializations
c
      npoit=npoin+(ntotv-ndofn*npoin)
      nnodw=nnode
      if(npoit.gt.npoin) nnodw=nnodt
c
c*** List of elements connected to the nodal points
c      
      call elepoi(nelem,nnodw,nnodt,npoit,lengp,lnods,svomt,
     .            p_nepoi,p_lelpo,1)
c
c*** Total number of nonzero coefficients
c
      nzeco=0
      itotv=0
      call sadmem(0,p_touch,nelem*nnodw,svomt)
      do ipoin=1,npoin                                      ! Main points
        kflag=0
        do idofn=1,ndofn
          itotv=itotv+1
          itotn=lpntn(itotv)                           
          if(itotn.gt.0) then                               ! Row of unknowns
            if(kflag.eq.0) then
              kflag=1
              call idenum(%val(p_nepoi),npoit,ipoin,nlist)
              ncoef=nlist*nnodw
            end if
            call slogze(ncoef,%val(p_touch))
            iadd1=p_lelpo+(ipoin-1)*lengp
            call spoias(%val(iadd1),iadd2)
            call nzecoe(nlist,nnode,nnodt,nelem,npoin,ndofn,
     .                  ntotv,ncoef,nzeco,%val(iadd2),
     .                  lnods,lpntn,%val(p_touch),nnodw,iblok)
          end if
        end do
      end do
      itotv=ndofn*npoin
      do ipoin=npoin+1,npoit                                ! Secondary points
        itotv=itotv+1
        itotn=lpntn(itotv) 
        if(itotn.gt.0) then                                 ! Row of unknowns
          call idenum(%val(p_nepoi),npoit,ipoin,nlist)
          ncoef=nlist*nnodw
          call slogze(ncoef,%val(p_touch))
          iadd1=p_lelpo+(ipoin-1)*lengp
          call spoias(%val(iadd1),iadd2)
          call nzecoe(nlist,nnode,nnodt,nelem,npoin,ndofn,
     .                ntotv,ncoef,nzeco,%val(iadd2),
     .                lnods,lpntn,%val(p_touch),nnodw,iblok)
        end if
      end do
c
c*** Allocate memory for A and the arrays of indexes. These are organized
c*** as follows: ROFFS(ITOTV) = coefficient of A where eqn. ITOTV starts,
c*** ACOLU(IZECO) = column of the IZECO coefficient of matrix A.
c
      lbyts=8*nzeco
      call sadmem(0,p_amatr,lbyts,speme)
      lbyts=lengp*(ntotv+1)
      call sadmem(0,p_roffs,lbyts,speme)
      call slvecz(ntotv+1,%val(p_roffs))
      lbyts=lengp*nzeco
      call sadmem(0,p_acolu,lbyts,speme)
      call slvecz(nzeco,%val(p_acolu))
c
c*** Construct the array of indexes
c
      itotv=0
      izeco=1
      do ipoin=1,npoin                                      ! Main points
        kflag=0
        do idofn=1,ndofn
          itotv=itotv+1
          itotn=lpntn(itotv) 
          if(itotn.gt.0) then                               ! Row of unknowns
            if(kflag.eq.0) then
              kflag=1
              call idenum(%val(p_nepoi),npoit,ipoin,nlist)
              ncoef=nlist*nnodw
            end if
            call slogze(ncoef,%val(p_touch))
            iadd1=p_lelpo+(ipoin-1)*lengp
            call spoias(%val(iadd1),iadd2)
            call arinde(nlist,nnode,nnodt,nelem,npoin,ndofn,
     .                  ntotv,ncoef,nzeco,%val(iadd2),
     .                  lnods,lpntn,%val(p_touch),
     .                  %val(p_roffs),%val(p_acolu),
     .                  izeco,itotv,nnodw,iblok)
          end if
        end do
      end do
      itotv=ndofn*npoin
      do ipoin=npoin+1,npoit                                ! Secondary points
        itotv=itotv+1
        itotn=lpntn(itotv) 
        if(itotn.gt.0) then                                 ! Row of unknowns
          call idenum(%val(p_nepoi),npoit,ipoin,nlist)
          ncoef=nlist*nnodw
          call slogze(ncoef,%val(p_touch))
          iadd1=p_lelpo+(ipoin-1)*lengp
          call spoias(%val(iadd1),iadd2)
          call arinde(nlist,nnode,nnodt,nelem,npoin,ndofn,
     .                ntotv,ncoef,nzeco,%val(iadd2),
     .                lnods,lpntn,%val(p_touch),
     .                %val(p_roffs),%val(p_acolu),
     .                izeco,itotv,nnodw,iblok)
        end if
      end do
c
c*** Deallocate volatile memory 
c
      svome = max(svome,svomt)
      call sadmem(2,p_touch,nelem*nnodw,svomt)
      call elepoi(nelem,nnodw,nnodt,npoit,lengp,lnods,svomt,
     .            p_nepoi,p_lelpo,2)
      
      end
